require 'gtk2'
require 'rexml/document'
require 'fileutils'
require 'pathname'

require 'gstudier'

class GStudyAppMainWindow < Gtk::Window
	
	def initialize(questions)
		super()
		set_title("Study application")
		@modified = false
		@questions = questions
		add_gui()
		load_study()

		signal_connect('delete_event') { |e, f|
			if @modified
				msgBox = Gtk::MessageDialog.new(self,
				      Gtk::Dialog::MODAL,
				      Gtk::MessageDialog::WARNING,
				      Gtk::MessageDialog::BUTTONS_YES_NO,
				      'Are you sure you want to quit without saving?')
				result = msgBox.run
				msgBox.destroy
				result != Gtk::Dialog::RESPONSE_YES
			else
				false
			end
		}
	end
	
	def add_gui()
		box1 = Gtk::VBox.new(false, 0)
		add(box1)

		add_menu(box1)
		add_toolbar(box1)
		
		@has_extra_cols = false
		if File.exists? 'plugins/extra_cols/extra_cols.rb'
			# display extra columns
			require 'plugins/extra_cols/extra_cols.rb'
			@has_extra_cols = true
		end

		col_types = [String]
		if @has_extra_cols
			col_types += ExtraCols.col_types()
		end
		@tvStore = Gtk::ListStore.new(*col_types)
		@tvStore.set_sort_column_id(0, Gtk::SortType::ASCENDING)
		@tv = Gtk::TreeView.new(@tvStore)
		@tv.selection.mode = Gtk::SELECTION_MULTIPLE
		@tv.signal_connect('row-activated') { |_tv, path, column|
			open_questionnary(selected_questionnary())
		}
		box1.add(@tv)
		desc_col = Gtk::TreeViewColumn.new("Questionnary", Gtk::CellRendererText.new, 'text' => 0)
		desc_col.resizable = true
		desc_col.sizing = Gtk::TreeViewColumn::FIXED
		desc_col.fixed_width = 160
		desc_col.sort_column_id = 0
		@tv.append_column(desc_col)

		if @has_extra_cols
			ExtraCols.col_names.each_with_index { |col_name, i|
				col = Gtk::TreeViewColumn.new(col_name, Gtk::CellRendererText.new, 'text' => i+1)
				col.resizable = true
				col.sizing = Gtk::TreeViewColumn::FIXED
				col.fixed_width = 160
				col.sort_column_id = i+1
				@tv.append_column(col)
			}
		end

	end

	def add_menu(box1)
		accelgroup = Gtk::AccelGroup.new
		add_accel_group(accelgroup)
		ifp = Gtk::ItemFactory.new(Gtk::ItemFactory::TYPE_MENU_BAR, "<main>", accelgroup)
		menu_items = [
			["/_File"],
			["/_File/_Save", "<StockItem>", "<control>S", Gtk::Stock::SAVE, Proc.new {save_study()}],
			*create_extra_items()
		]
		menu_items << ["/_File/_Quit", "<StockItem>", "<control>Q", Gtk::Stock::QUIT, Proc.new {Gtk.main_quit}]
		ifp.create_items(menu_items)
		box1.pack_start(ifp.get_widget("<main>"), false, true, 0)
	end

	def add_toolbar(box1)
		@toolbar = Gtk::Toolbar.new
		@toolbar.append(Gtk::Stock::SAVE, "Save questionnary data") do
			save_study()
		end
		@toolbar.append(Gtk::Stock::ADD, "Add a questionnary") do
			add_questionnary()
		end
		@toolbar.append(Gtk::Stock::PROPERTIES, "Edit selected questionnary") do
			if selected_questionnary()
				open_questionnary(selected_questionnary())
			else
				msgBox = Gtk::MessageDialog.new(self,
				      Gtk::Dialog::MODAL,
				      Gtk::MessageDialog::WARNING,
				      Gtk::MessageDialog::BUTTONS_OK,
				      'You must select a questionnary to edit!')
				msgResult = msgBox.run
				msgBox.destroy
			end
		end
		@toolbar.append(Gtk::Stock::DELETE, "Delete selected questionnary") do
			if selected_questionnary()
				delete_questionnary(selected_questionnary())
			else
				msgBox = Gtk::MessageDialog.new(self,
				      Gtk::Dialog::MODAL,
				      Gtk::MessageDialog::WARNING,
				      Gtk::MessageDialog::BUTTONS_OK,
				      'You must select a questionnary to delete!')
				msgResult = msgBox.run
				msgBox.destroy
			end
		end
		box1.pack_start(@toolbar, false)
	end

	def create_extra_items()
		result = []
		Dir['plugins/export/*.rb'].each { |plugin_file|
			# export menu item
			require plugin_file
			base_name = Pathname.new(plugin_file).basename.to_s
			class_name = base_name.sub(/#{Regexp.escape(Pathname.new(plugin_file).extname.to_s)}$/, '').capitalize
			obj = eval(class_name + ".new")
			export_proc = Proc.new { |f, tv, export_selected| obj.export(@doc, tv, @questions, f, export_selected) }
			result <<  ["/_File/_Export #{obj.export_desc()}", "<StockItem>", "<control>E", Gtk::Stock::CONVERT, Proc.new { ask_saving(obj.extension(), export_proc, obj.support_export_selection()) }]
		}
		result
	end

	class Gtk::TreeView
		def mult_has_selection()
			selection.selected_each() { |model, path, iter|
				return true
			}
			false
		end
	end

	def ask_saving(f_extension, saving_proc, support_export_selection)
		export_selected_if_possible = true
		if support_export_selection &&
			@tv.mult_has_selection()
			dialog = Gtk::MessageDialog.new(nil, Gtk::MessageDialog::MODAL,
			Gtk::MessageDialog::QUESTION, Gtk::MessageDialog::BUTTONS_YES_NO, "Click yes if you want to export only the selected questionnaries. No to export all of them.")
			resp = dialog.run
			dialog.destroy
			if resp == Gtk::MessageDialog::RESPONSE_NO
				export_selected_if_possible = false
			end
		end
# 		if Gtk::MINOR_VERSION < 4
			fs = Gtk::FileSelection.new("Please choose the file to export to.")
			fs.ok_button.signal_connect('clicked') {
				f_name = fs.filename
				if f_name !~ /#{Regexp.escape(f_extension)}$/
					# user didn't enter the file extension
					# add it
					f_name += ".#{f_extension}"
				end
				if FileTest.directory?(f_name)
					fs.show_all
				else
					begin
						File.open(f_name, 'w') { |f|
							saving_proc[f, @tv, export_selected_if_possible]
						}
						fs.destroy
					rescue => e
						display_exception(e)
						fs.show_all
					end
				end
			}
			fs.cancel_button.signal_connect('clicked') {
				fs.destroy
			}
			fs.show_all
# 		else
# 			fs = Gtk::FileChooserDialog.new("Please choose the file to export to.", self, Gtk::FileChooser::ACTION_SAVE, nil,
# 				     [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_ACCEPT],
# 				     [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL]
# 			)
# 			filter_rb = Gtk::FileFilter.new
# 			filter_rb.name = "CSV file"
# 			filter_rb.add_pattern("*.csv")
# 			fs.add_filter(filter_rb)
# 			if fs.run == Gtk::Dialog::RESPONSE_ACCEPT
# 				f_name = fs.filename
# 				if f_name !~ /#{Regexp.escape(f_extension)}$/
# 					# user didn't enter the file extension
# 					# add it
# 					f_name += ".#{f_extension}"
# 				end
# 				if !FileTest.directory?(f_name)
# 					begin
# 						puts "SAVING! #{f_name}"
# 						File.open(f_name, 'w') { |f|
# 							saving_proc[f, @tv, export_selected_if_possible]
# 						}
# 					rescue => e
# 						display_exception(e)
# 					end
# 				end
# 			end
# 			fs.destroy
# 		end
	end

	def load_study()
		f = File.open('study.sty')
		@doc = REXML::Document.new f
		f.close
		refresh_questionnary_list()
	end

	def refresh_questionnary_list()
		@tvStore.clear
		@doc.root.elements.each { |elt|
			treeiter = @tvStore.append
			treeiter[0] = elt.attributes["number"]

			if @has_extra_cols
				1.upto(ExtraCols.col_names.size) { |i|
					treeiter[i] = ExtraCols.col_value(@questions, @doc, elt.attributes["number"], i)
				}
			end
		}
	end

	def save_study()
		if File.exists? 'study.sty'
			FileUtils.cp('study.sty', 'study.sty.bak')
		end
		f = File.open('study.sty', 'w')
		@doc.write(f, 2)
		f.close
		@modified = false
	end

	def selected_questionnary()
		result = nil
		@tv.selection.selected_each { |model, path, iter|
			quest_id = iter[0]
			return @doc.root.elements.to_a.find { |elt| elt.attributes["number"] == quest_id}
		}
		nil
	end

	def add_questionnary()
		gstudier = GStudierDialog.new(@questions, nil, last_questionnary_id()+1)
		result = display_questionnary(gstudier, nil)
		if result == Gtk::Dialog::RESPONSE_OK 
			@modified = true
			question_number = gstudier.question_no()

			questionnary_node = REXML::Element.new "questionnary"
			questionnary_node.attributes["number"] = question_number.to_s
			questionnary_node.add_element gstudier.entries_xml()
			@doc.root.add_element questionnary_node
			refresh_questionnary_list()
			gstudier.destroy 
		end
	end

	def delete_questionnary(questionnary_node)
		msgBox = Gtk::MessageDialog.new(self,
		      Gtk::Dialog::MODAL,
		      Gtk::MessageDialog::WARNING,
		      Gtk::MessageDialog::BUTTONS_YES_NO,
		      'Are you sure you want to delete this questionnary?')
		msgResult = msgBox.run
		msgBox.destroy
		if msgResult == Gtk::Dialog::RESPONSE_YES
			questionnary_node.parent.delete_element questionnary_node
			@modified = true
			refresh_questionnary_list()
		end
	end

	def open_questionnary(questionnary_node)
		# get the matching questionnary XML
		gstudier = GStudierDialog.new(@questions, questionnary_node)
		cur_quest_id = questionnary_node.attributes['number'].to_i
		cur_quest_id
		result = display_questionnary(gstudier, cur_quest_id)
		if result == Gtk::Dialog::RESPONSE_OK 
			question_number = gstudier.question_no()
			@doc.root.delete_element questionnary_node

			new_questionnary_node = REXML::Element.new "questionnary"
			new_questionnary_node.attributes["number"] = question_number.to_s
			new_questionnary_node.add_element gstudier.entries_xml()
			@doc.root.add_element new_questionnary_node
			@modified = true
			refresh_questionnary_list()

			gstudier.destroy 
		end
	end

	def display_questionnary(gstudier, cur_questionnary_id)
		gstudier.set_default_size(700, 500)
		result = gstudier.run
		while result == Gtk::Dialog::RESPONSE_DELETE_EVENT ||
				result == Gtk::Dialog::RESPONSE_CANCEL ||
				gstudier.question_no() <= 0 ||
				questionnary_id_invalid(gstudier.question_no(), cur_questionnary_id)

			question_number = gstudier.question_no()
			if result == Gtk::Dialog::RESPONSE_OK
				if question_number <= 0
					msgBox = Gtk::MessageDialog.new(gstudier,
					      Gtk::Dialog::MODAL,
					      Gtk::MessageDialog::WARNING,
					      Gtk::MessageDialog::BUTTONS_OK,
					      'For the questionnary number, you must enter an integer superior to 0')
					msgResult = msgBox.run
					msgBox.destroy
					# reshow the dialog
					result = gstudier.run
				elsif questionnary_id_invalid(gstudier.question_no(), cur_questionnary_id)
					msgBox = Gtk::MessageDialog.new(gstudier,
					      Gtk::Dialog::MODAL,
					      Gtk::MessageDialog::WARNING,
					      Gtk::MessageDialog::BUTTONS_OK,
					      'The questionnary number you gave is already used for another questionnary. Please choose another questionnary number.')
					msgResult = msgBox.run
					msgBox.destroy
					# reshow the dialog
					result = gstudier.run
				end
			else
				msgBox = Gtk::MessageDialog.new(gstudier,
				      Gtk::Dialog::MODAL,
				      Gtk::MessageDialog::WARNING,
				      Gtk::MessageDialog::BUTTONS_YES_NO,
				      'Are you sure you want to cancel editing this questionnary?')
				msgResult = msgBox.run
				msgBox.destroy
				if msgResult == Gtk::Dialog::RESPONSE_YES
					# stop adding knowledge.
					gstudier.destroy 
					return
				else
					# reshow the dialog
					result = gstudier.run
				end
			end
		end
		result
	end

	# the user chose the chosen_question_id for questionnary id
	# while editing an existing questionnary or adding a new
	# questionnary.
	# if it's editing, cur_question_id represents the Fixnum id
	# of the edited questionnary. if it's adding, cur_question_id
	# is nil.
	# this method checks whether the user chose a valid questionnary_id
	# this id must not be the id of an existing questionnary, but
	# if it's the id of the current questionnary in the case of editing,
	# it's ok.
	def questionnary_id_invalid(chosen_question_id, cur_question_id)
		if chosen_question_id == cur_question_id
			return false # all OK.
		end
		@doc.root.each_element { |elt|
			if elt.attributes['number'].to_i == chosen_question_id
				# the user chose a questionnary id
				# that already existed!
				return true
			end
		}
		return false
	end

	def last_questionnary_id()
		max = 0
		@doc.root.each_element { |elt|
			cur_val = elt.attributes["number"].to_i
			max = cur_val if cur_val > max
		}
		max
	end
	
end

def display_exception(info)
	traceWithLines = info.backtrace.map {|entry|
		file, lineNo = entry.split(':')
		lineNo = lineNo.to_i
		line = ''
		begin
		File.open(file) {|f|
			line = f.readlines[lineNo-1].chomp.strip
		}
		rescue
		end
		entry + " (#{line})"
	}.join "\n"
	msgBox = Gtk::MessageDialog.new(nil,
	      Gtk::Dialog::MODAL |
	      Gtk::Dialog::DESTROY_WITH_PARENT,
	      Gtk::MessageDialog::ERROR,
	      Gtk::MessageDialog::BUTTONS_OK,
		%Q!Crash.. send a shot of this to emmanuel :\n\n#{info.message}\n#{traceWithLines}!)
	msgBox.run
	msgBox.destroy
end

if $0 == __FILE__
	Gtk.init
	
	filename = ARGV[0] || fail("must give filename of question data as first arg")
	begin
		questions = nil
		File.open(filename) { |f|
			questions = Studier::parse_questions(f)
		}

		main = GStudyAppMainWindow.new(questions)
		main.set_default_size(700, 500)
		
		main.show_all
		
		main.signal_connect('delete_event') do
		  Gtk::main_quit
		end
		
		Gtk.main
	
	rescue Exception => info
		display_exception(info)
	end
end
